from PyQt6.QtWidgets import (
    QHBoxLayout,
    QLineEdit,
    QPushButton,
    QStyle,
    QWidget,
)
from PyQt6.QtCore import pyqtSignal
from karoto.flow_layout import FlowLayout


class Tag(QPushButton):
    deleted = pyqtSignal()

    def __init__(self, name: str, count: int = None, cross=True) -> None:
        super().__init__()

        self.name = name
        self.setText(name if count is None else f"{name} ({count})")

        if cross:
            pixmapi = QStyle.StandardPixmap.SP_DialogCloseButton
            icon = self.style().standardIcon(pixmapi)
            self.setIcon(icon)

        self.clicked.connect(self.deleted)


class TagWizzard(QWidget):
    created_tag = pyqtSignal(str)

    def __init__(self) -> None:
        super().__init__()

        self.init_ui()

    def init_ui(self) -> None:
        self.plus_button = QPushButton("+")
        self.plus_button.clicked.connect(self.toggle)
        self.plus_button.setProperty("class", "single_char_button")

        self.name_edit = QLineEdit()
        self.name_edit.returnPressed.connect(self.ready)
        self.name_edit.setHidden(True)

        self.confirm_button = QPushButton(">")
        self.confirm_button.clicked.connect(self.ready)
        self.confirm_button.setHidden(True)
        self.confirm_button.setProperty("class", "single_char_button")

        self.layout_ = QHBoxLayout()
        self.layout_.addWidget(self.plus_button)
        self.layout_.addWidget(self.name_edit)
        self.layout_.addWidget(self.confirm_button)
        self.layout_.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout_)

    def toggle(self) -> None:
        self.plus_button.setHidden(not self.plus_button.isHidden())
        self.name_edit.setHidden(not self.name_edit.isHidden())
        self.confirm_button.setHidden(not self.confirm_button.isHidden())

        self.name_edit.setFocus()

    def ready(self) -> None:
        name = self.name_edit.text()
        if name == "":
            self.toggle()
            return
        self.created_tag.emit(name)


class Tags(QWidget):
    deleted_tag = pyqtSignal(str)
    created_tag = pyqtSignal(str)

    def __init__(self, tags: list[str]) -> None:
        super().__init__()
        self.tags = tags
        self.init_ui()

    def init_ui(self) -> None:
        self.layout_ = FlowLayout()
        for name in self.tags:
            t = Tag(name=name)
            t.deleted.connect(
                    lambda x=name: self.deleted_tag.emit(x)
            )
            self.layout_.addWidget(t)
        self.wizzard = TagWizzard()
        self.wizzard.created_tag.connect(self.created_tag)
        self.layout_.addWidget(self.wizzard)

        self.setLayout(self.layout_)


class TagFilter(QWidget):
    changed = pyqtSignal()

    def __init__(self) -> None:
        super().__init__()

        self.all_tags = []
        self.filter_tags = []

        self.init_ui()

    def init_ui(self) -> None:
        self.layout_ = FlowLayout()
        self.setLayout(self.layout_)

    def show_tags(self, add_another=False) -> None:
        self.clear_layout()

        for filter_tag in self.filter_tags:
            t = Tag(name=filter_tag)
            t.deleted.connect(lambda x=filter_tag: self.remove_filter(x))
            self.layout_.addWidget(t)

        if self.filter_tags != [] and not add_another:
            add_button = QPushButton("+")
            add_button.setProperty("class", "single_char_button")
            add_button.clicked.connect(
                lambda: self.show_tags(add_another=True)
            )
            self.layout_.addWidget(add_button)
            return

        for name, count in self.all_tags:
            if name in self.filter_tags:
                continue

            t = Tag(name=name, count=count, cross=False)
            t.deleted.connect(lambda x=name: self.add_filter(x))
            self.layout_.addWidget(t)

    def clear_layout(self) -> None:
        while not self.layout_.isEmpty():
            del_item = self.layout_.itemAt(0)
            self.layout_.removeItem(del_item)
            del_item.widget().deleteLater()

    def update_all_tags(self, all_tags: list[tuple[str, int]]) -> None:
        self.all_tags = all_tags
        self.show_tags()

    def add_filter(self, name: str) -> None:
        self.filter_tags.append(name)
        self.changed.emit()

    def remove_filter(self, name: str) -> None:
        self.filter_tags.remove(name)
        self.changed.emit()



import sys
from PyQt6.QtWidgets import (
    QApplication,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QScroller,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)
from PyQt6.QtCore import pyqtSignal, Qt
from karoto.backend import CartItem
from karoto.clickable_label import ClickableLabel
from karoto.base_feed import BaseFeed


class ShoppingFeedItemDetails(QWidget):
    item_changed = pyqtSignal()
    warn_message = pyqtSignal(str)
    needs_reload = pyqtSignal()

    def __init__(self, item: CartItem) -> None:
        super().__init__()
        self.item = item

        self.init_ui()

    def init_ui(self) -> None:
        self.hide_button = QPushButton("Hide temporary")
        self.hide_button.clicked.connect(self.hide_temporary)

        self.minus_button = QPushButton("-")
        self.minus_button.clicked.connect(self.decrease)
        self.minus_button.setProperty("class", "single_char_button")

        self.missing_edit = QLineEdit("ERROR")
        self.missing_edit.editingFinished.connect(self.update_missing)
        self.missing_edit.setSizePolicy(
            QSizePolicy.Policy.Minimum,
            self.missing_edit.sizePolicy().verticalPolicy(),
        )
        self.missing_edit.setProperty("class", "short_line_edit")

        self.wanted_label = QLabel(f"/ {self.item.wanted}")

        self.plus_button = QPushButton("+")
        self.plus_button.clicked.connect(self.increase)
        self.plus_button.setProperty("class", "single_char_button")

        self.layout_ = QHBoxLayout()
        self.layout_.addStretch()
        self.layout_.addWidget(self.hide_button)
        self.layout_.addWidget(self.minus_button)
        self.layout_.addWidget(self.missing_edit)
        self.layout_.addWidget(self.wanted_label)
        self.layout_.addWidget(self.plus_button)
        self.setLayout(self.layout_)

    def reload(self) -> None:
        self.missing_edit.setText(str(self.item.missing))
        self.plus_button.setEnabled(self.item.in_stock > 0)

    def update_missing(self) -> None:
        try:
            new = float(self.missing_edit.text())

            if self.item.missing == new:
                return

            if new > self.item.wanted:
                self.warn_message.emit(
                    "You can't go higher than the wanted amount ("
                    f"{self.item.wanted} {self.item.unit})",
                )
                # TODO: Maybe introduce something like only once wanted that
                #       allows this? Is it worth it?

            if new < 0:
                self.warn_message.emit("Negative numbers not supported!")
                self.missing_edit.setText(str(self.item.missing))
                return

            self.item.missing = new

        except ValueError as e:
            self.warn_message.emit(str(e))
            self.missing_edit.setText(str(self.item.missing))
            return

        self.item_changed.emit()

    def increase(self) -> None:
        self.item.missing += 1
        self.item_changed.emit()

    def decrease(self) -> None:
        self.item.missing -= 1
        self.item_changed.emit()

    def hide_temporary(self) -> None:
        self.item.hidden_temporary = True
        self.needs_reload.emit()


class ShoppingFeedItem(QGroupBox):
    item_changed = pyqtSignal()
    bought_all = pyqtSignal()
    warn_message = pyqtSignal(str)
    needs_reload = pyqtSignal()

    def __init__(self, item: CartItem) -> None:
        super().__init__()
        self.item = item

        self.init_ui()

    def init_ui(self) -> None:
        self.name_label = ClickableLabel("ERROR")
        self.name_label.setWordWrap(True)
        self.name_label.setSizePolicy(
            QSizePolicy.Policy.Expanding,
            self.name_label.sizePolicy().verticalPolicy(),
        )
        self.name_label.clicked.connect(self.toggle_details)

        self.more_button = QPushButton("\u2B0E")
        self.more_button.setProperty("class", "single_char_button")
        self.more_button.clicked.connect(self.toggle_details)
        self.more_button.setSizePolicy(
            QSizePolicy.Policy.Minimum,
            self.more_button.sizePolicy().verticalPolicy(),
        )

        self.bought_button = QPushButton("bought")
        self.bought_button.clicked.connect(self.buy_all)
        self.bought_button.setSizePolicy(
            QSizePolicy.Policy.Minimum,
            self.bought_button.sizePolicy().verticalPolicy(),
        )

        self.details = None  # see self._init_details()

        self.layout_ = QVBoxLayout()
        self.top_layout = QHBoxLayout()
        self.top_layout.addWidget(self.name_label)
        self.top_layout.addWidget(self.more_button)
        self.top_layout.addWidget(self.bought_button)
        self.layout_.addLayout(self.top_layout)
        # do not simplify self.layout_ as details get added later
        self.setLayout(self.layout_)

        self.reload()

    def _init_details(self) -> None:
        """details get initialized when needed only to speed up the app"""
        self.details = ShoppingFeedItemDetails(self.item)
        self.details.item_changed.connect(self.item_changed)
        self.details.item_changed.connect(self.check_bought_all)
        self.details.item_changed.connect(self.reload)
        self.details.needs_reload.connect(self.needs_reload)
        self.details.warn_message.connect(self.warn_message)
        self.details.setHidden(True)
        self.layout_.addWidget(self.details)

    def check_bought_all(self) -> None:
        if not self.item.needs_restock:
            self.bought_all.emit()

    def reload(self) -> None:
        self.name_label.setText(
            f"{self.item.name}: {self.item.missing} {self.item.unit}"
        )

        if self.details is None:
            return

        if self.details.isHidden():
            self.more_button.setText("\u2B0E")
        else:
            self.more_button.setText("\u2B11")

        self.details.reload()

    def buy_all(self) -> None:
        self.item.bought_all()
        self.bought_all.emit()
        self.item_changed.emit()

    def toggle_details(self) -> None:
        if self.details is None:
            self._init_details()

        self.details.setHidden(not self.details.isHidden())
        self.reload()


class ShoppingFeed(BaseFeed):
    item_changed = pyqtSignal(CartItem)
    warn_message = pyqtSignal(str)
    needs_reload = pyqtSignal()

    def init_ui(self) -> None:
        self.scroll_widget = QWidget()
        self.scroll_layout = QVBoxLayout()
        self.item_layout = QVBoxLayout()
        self.scroll_layout.addLayout(self.item_layout)
        self.scroll_layout.addStretch(1)

        self.scroll_widget.setLayout(self.scroll_layout)
        self.setWidget(self.scroll_widget)

        self.setWidgetResizable(True)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
        self.setHorizontalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        )
        QScroller.grabGesture(
            self.viewport(), QScroller.ScrollerGestureType.TouchGesture
        )

    def _create_ui_item(self, item: CartItem) -> ShoppingFeedItem:
        si = ShoppingFeedItem(item)
        si.item_changed.connect(lambda x=item: self.item_changed.emit(x))
        si.bought_all.connect(lambda x=item: self.remove_item(x))
        si.needs_reload.connect(self.needs_reload)
        si.warn_message.connect(self.warn_message)
        return si


def main() -> None:
    app = QApplication(sys.argv)
    cart_items = [
        CartItem(
            name="Kräutersalz",
            wanted=1,
            in_stock=0,
            unit="Stk",
            only_once=False,
        ),
        CartItem(
            name="Kartoffeln",
            wanted=2.5,
            in_stock=1,
            unit="kg",
            only_once=False,
        ),
    ]
    for i in range(40):
        i += 1
        cart_items.append(
            CartItem(
                name=f"Item {i}",
                wanted=2 * i,
                in_stock=i,
                unit="Stk",
                only_once=False,
            )
        )

    shopping_feed = ShoppingFeed(cart_items)
    shopping_feed.item_changed.connect(print)
    shopping_feed.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



from PyQt6.QtWidgets import (
    QScrollArea,
    QWidget,
)
from karoto.backend import CartItem


class BaseFeed(QScrollArea):  # TODO: Make ABC?!
    def __init__(self):
        super().__init__()

        self.cart_items = list()
        self.ui_items = list()
        self._n_items_scrollbar_range = None

        self.init_ui()

        self.verticalScrollBar().valueChanged.connect(
            self.on_scrollbar_value_changed
        )
        self.verticalScrollBar().rangeChanged.connect(
            self.on_scrollbar_range_changed
        )

    # @abstractmethod
    def init_ui(self) -> None:
        raise NotImplementedError()

    # @abstractmethod
    def _create_ui_item(self, item: CartItem) -> QWidget:
        raise NotImplementedError()

    def on_scrollbar_value_changed(self, value: int) -> None:
        if self._range_didnt_change_yet:
            # This gets triggered too early and we'd load too many items
            return

        if value > (self.verticalScrollBar().maximum()
                    - self._n_items_scrollbar_range):
            self._range_didnt_change_yet = True
            self.load_n_items()

    def on_scrollbar_range_changed(self, min_: int, max_: int) -> None:
        self._range_didnt_change_yet = False

        if max_ == 0:
            return  # on start the range is always 0

        if self._n_items_scrollbar_range is None:  # only once executed
            self._n_items_scrollbar_range = max_

            # Now we know the scrollbar range so we can load another n items
            self.load_n_items()

    def clear_layout(self) -> None:
        while not self.item_layout.isEmpty():
            del_item = self.item_layout.itemAt(0)
            self.item_layout.removeItem(del_item)
            del_item.widget().deleteLater()

    def show_items(self, items: list[CartItem]) -> None:
        self.cart_items = items
        self.reload_items()

    def reload_items(self) -> None:
        self.clear_layout()
        self.load_n_items()

    def load_n_items(self) -> None:
        """only load a fraction of all items to lower loading times"""
        num_items_to_load = 20

        old_item_count = self.item_layout.count()
        if old_item_count >= len(self.cart_items):
            return

        for count, item in enumerate(self.cart_items):
            if count < old_item_count:
                continue  # item was already loaded
            if count >= old_item_count + num_items_to_load:
                break  # we loaded n items and are ready

            new_item = self._create_ui_item(item)
            self.item_layout.addWidget(new_item)

    def remove_item(self, item: CartItem) -> None:
        self.cart_items.remove(item)
        self.reload_items()



#!/usr/bin/python3

import sys
from argparse import ArgumentParser
from PyQt6.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QLineEdit,
    QPushButton,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)
from PyQt6.QtGui import QIcon, QImage, QPixmap
from karoto.backend import Backend, CartItem, DuplicateError
from karoto.storage_feed import StorageFeed
from karoto.shopping_feed import ShoppingFeed
from karoto.error_bar import ErrorBar
from karoto.tags import TagFilter
from karoto import style


class MainWindow(QWidget):
    def __init__(self, backend: Backend):
        super().__init__()

        self.backend = backend
        self.init_ui()

    def init_ui(self) -> None:
        self.setStyleSheet(style.default)

        topbar = QHBoxLayout()

        self.menu_button = QPushButton()
        self.menu_button.clicked.connect(self.toggle_view)
        self.menu_button.setObjectName("menu_button")
        self.menu_button.setProperty("class", "top_bar")
        topbar.addWidget(self.menu_button)

        self.filter_button = QPushButton()
        self.filter_button.clicked.connect(self.toggle_filter)
        self.filter_button.setProperty("class", "top_bar")
        topbar.addWidget(self.filter_button)

        self.searchfield = QLineEdit()
        self.searchfield.textChanged.connect(lambda: self.refresh())
        self.searchfield.setPlaceholderText("\U0001F50D Search...")
        self.searchfield.setObjectName("search_field")
        self.searchfield.setProperty("class", "top_bar")
        topbar.addWidget(self.searchfield)

        self.tag_filter = TagFilter()
        self.tag_filter.changed.connect(self.refresh)
        self.tag_filter.setHidden(False)
        self.toggle_filter()

        self.error_bar = ErrorBar()
        self.error_bar.setObjectName("error_bar")

        self.shopping_feed = ShoppingFeed()
        self.storage_feed = StorageFeed()

        self.shopping_feed.item_changed.connect(self.on_item_changed)
        self.shopping_feed.warn_message.connect(self.error_bar.show_warning)
        self.shopping_feed.needs_reload.connect(self.refresh)
        self.storage_feed.item_changed.connect(self.on_item_changed)
        self.storage_feed.item_deleted.connect(self.delete)
        self.storage_feed.new_item.connect(self.on_new_item)
        self.storage_feed.name_changed.connect(self.on_name_changed)
        self.storage_feed.warn_message.connect(self.error_bar.show_warning)

        self.stack = QStackedWidget()
        self.stack.addWidget(self.shopping_feed)
        self.stack.addWidget(self.storage_feed)

        vbox = QVBoxLayout()
        vbox.addLayout(topbar)
        vbox.addWidget(self.tag_filter)
        vbox.addWidget(self.error_bar)
        vbox.addWidget(self.stack)

        vbox.setContentsMargins(0, 0, 0, 0)

        self.setLayout(vbox)

        self.refresh()
        self.show()

    @property
    def is_in_storage_mode(self) -> bool:
        return self.stack.currentWidget() == self.storage_feed

    def refresh(self) -> None:
        self.tag_filter.update_all_tags(
            self.backend.get_tags(
                search=self.searchterm,
                filter_tags=self.tag_filter.filter_tags,
                storage_mode=self.is_in_storage_mode,
            )
        )

        if self.is_in_storage_mode:
            self.storage_feed.show_items(
                self.backend.get_items_in_stock(
                    search=self.searchterm,
                    filter_tags=self.tag_filter.filter_tags,
                ),
            )
            menu_icon_svg = style.menu_home_icon_svg
        else:
            self.shopping_feed.show_items(
                self.backend.get_items_to_buy(
                    search=self.searchterm,
                    filter_tags=self.tag_filter.filter_tags,
                ),
            )
            menu_icon_svg = style.menu_cart_icon_svg

        menu_icon = QIcon(
            QPixmap(QImage.fromData(menu_icon_svg))
        )
        self.menu_button.setIcon(menu_icon)

    @property
    def searchterm(self) -> str:
        return self.searchfield.text()

    def on_new_item(self, item: CartItem) -> None:
        try:
            self.backend.add_item(item)
        except DuplicateError:
            self.error_bar.show_warning(
                f'Item "{item.name}" already exists. Aborting...'
            )
            return

        self.backend.save_list()
        self.refresh()

    def on_item_changed(self, item: CartItem) -> None:
        if item.is_dead:
            self.delete(item)
        else:
            self.backend.save_list()

    def on_name_changed(self, item: CartItem, new_name: str) -> None:
        if self.backend.get_item_by_name(new_name) is not None:
            self.error_bar.show_warning(
                f'Tried to rename "{item.name}" to "{new_name}"'
                " which already exists. Aborting..."
            )
            return
        item.name = new_name

    def delete(self, item: CartItem) -> None:
        self.backend.delete(item)
        self.backend.save_list()
        self.refresh()

    def toggle_view(self) -> None:
        if self.stack.currentWidget() == self.storage_feed:
            self.stack.setCurrentWidget(self.shopping_feed)
        else:
            self.stack.setCurrentWidget(self.storage_feed)
        self.refresh()

    def toggle_filter(self) -> None:
        self.tag_filter.setHidden(not self.tag_filter.isHidden())

        if self.tag_filter.isHidden():
            filter_icon_svg = style.filter_plus_icon_svg
        else:
            filter_icon_svg = style.filter_minus_icon_svg

        filter_icon = QIcon(
            QPixmap(QImage.fromData(filter_icon_svg))
        )
        self.filter_button.setIcon(filter_icon)


def main():
    parser = ArgumentParser()
    parser.add_argument(
        "--list", "-l",
        type=str,
        help="Use an alternative list by name (without extension)",
        default="default",
    )
    parser.add_argument(
        "--list-file",
        type=str,
        help="Use an alternative list file",
        default=None,
    )
    args = parser.parse_args()
    app = QApplication(sys.argv)
    backend = Backend(list_name=args.list, list_file=args.list_file)
    gui = MainWindow(backend=backend)  # noqa: F841
    return app.exec()


if __name__ == "__main__":
    sys.exit(main())



import sys
from PyQt6.QtWidgets import (
    QApplication,
    QCheckBox,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QScroller,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)
from PyQt6.QtCore import pyqtSignal, Qt
from karoto.backend import CartItem
from karoto.clickable_label import ClickableLabel
from karoto.base_feed import BaseFeed
from karoto.tags import Tags


class StorageFeedItemDetails(QWidget):
    item_saved = pyqtSignal()
    item_deleted = pyqtSignal()
    name_changed = pyqtSignal(str)

    def __init__(self, item: CartItem, is_for_wizzard=False) -> None:
        super().__init__()
        self.item = item
        self.last_saved_tags = item.tags  # TODO: Hacky...
        self.is_for_wizzard = is_for_wizzard

        self.init_ui()

    def init_ui(self) -> None:
        self.layout_ = QGridLayout()
        self.setLayout(self.layout_)

        self.name_label = QLabel("Name:")
        self.name_edit = QLineEdit("ERROR")
        self.name_edit.textChanged.connect(self.check_changes)

        self.in_stock_label = QLabel("In stock:")
        self.in_stock_edit = QLineEdit("ERROR")
        self.in_stock_edit.textChanged.connect(self.check_changes)

        self.wanted_label = QLabel("Wanted:")
        self.wanted_edit = QLineEdit("ERROR")
        self.wanted_edit.textChanged.connect(self.check_changes)

        self.unit_label = QLabel("Unit:")
        self.unit_edit = QLineEdit("ERROR")
        self.unit_edit.textChanged.connect(self.check_changes)

        self.only_once_label = QLabel("Only Once:")
        self.only_once_check = QCheckBox()
        self.only_once_check.toggled.connect(self.check_changes)

        self.tags_label = QLabel("Tags:")
        self.init_tags()

        self.save_button = QPushButton(
            "Create Item" if self.is_for_wizzard else "Save"
        )
        self.save_button.setEnabled(False)
        self.save_button.clicked.connect(self.save_changes)

        if not self.is_for_wizzard:
            self.delete_button = QPushButton("Delete")
            self.delete_button.clicked.connect(self.item_deleted)

        self.layout_.addWidget(self.name_label, 0, 0)
        self.layout_.addWidget(self.name_edit, 0, 1)
        self.layout_.addWidget(self.in_stock_label, 1, 0)
        self.layout_.addWidget(self.in_stock_edit, 1, 1)
        self.layout_.addWidget(self.wanted_label, 2, 0)
        self.layout_.addWidget(self.wanted_edit, 2, 1)
        self.layout_.addWidget(self.unit_label, 3, 0)
        self.layout_.addWidget(self.unit_edit, 3, 1)
        self.layout_.addWidget(self.only_once_label, 4, 0)
        self.layout_.addWidget(self.only_once_check, 4, 1)
        self.layout_.addWidget(self.tags_label, 5, 0)
        self.layout_.addWidget(self.save_button, 6, 1)
        if not self.is_for_wizzard:
            self.layout_.addWidget(self.delete_button, 7, 1)

        self.reload()

    def init_tags(self) -> None:
        self.tags_widget = Tags(self.item.tags)
        self.layout_.addWidget(self.tags_widget, 5, 1)
        self.tags_widget.created_tag.connect(self.create_tag)
        self.tags_widget.created_tag.connect(self.check_changes)
        self.tags_widget.deleted_tag.connect(self.delete_tag)
        self.tags_widget.deleted_tag.connect(self.check_changes)

    def reload(self) -> None:
        self.name_edit.setText(self.item.name)
        self.in_stock_edit.setText(str(self.item.in_stock))
        self.wanted_edit.setText(str(self.item.wanted))
        self.unit_edit.setText(self.item.unit)
        self.only_once_check.setChecked(self.item.only_once)
        self.reload_tags()

    def reload_tags(self):
        self.layout_.removeWidget(self.tags_widget)
        self.tags_widget.deleteLater()
        self.init_tags()

    def create_tag(self, name) -> None:
        self.item.add_tag(name)
        self.reload_tags()

    def delete_tag(self, name) -> None:
        self.item.remove_tag(name)
        self.reload_tags()

    def check_changes(self) -> None:
        found_changes = False

        if self.item.name != self.name_edit.text():
            found_changes = True

        try:
            if self.item.in_stock != float(self.in_stock_edit.text()):
                found_changes = True
        except ValueError:
            pass

        try:
            if self.item.wanted != float(self.wanted_edit.text()):
                found_changes = True
        except ValueError:
            pass

        if self.item.unit != self.unit_edit.text():
            found_changes = True

        if self.item.only_once != self.only_once_check.isChecked():
            found_changes = True

        if self.last_saved_tags != self.tags_widget.tags:
            found_changes = True

        self.save_button.setEnabled(found_changes)

    def save_changes(self) -> None:
        new_name = self.name_edit.text()
        if self.is_for_wizzard:
            self.item.name = new_name
        elif self.item.name != new_name:
            self.name_changed.emit(new_name)

        self.item.in_stock = self.in_stock_edit.text()
        self.item.wanted = self.wanted_edit.text()
        self.item.unit = self.unit_edit.text()
        self.item.only_once = self.only_once_check.isChecked()
        self.last_saved_tags = self.item.tags

        self.item_saved.emit()

        self.save_button.setEnabled(False)


class StorageFeedItem(QGroupBox):
    item_changed = pyqtSignal()
    item_deleted = pyqtSignal()
    name_changed = pyqtSignal(str)
    warn_message = pyqtSignal(str)

    def __init__(self, item: CartItem) -> None:
        super().__init__()
        self.item = item

        self.init_ui()

    def init_ui(self) -> None:
        self.details = None  # see self._init_details()

        self.layout_ = QVBoxLayout()
        self.top_layout = QHBoxLayout()
        self.layout_.addLayout(self.top_layout)
        # details get added to self.layout_ too so don't simplify the layout
        self.setLayout(self.layout_)

        self.name_label = ClickableLabel("ERROR")
        self.name_label.setWordWrap(True)
        self.name_label.clicked.connect(self.toggle_details)
        self.name_label.setSizePolicy(
            QSizePolicy.Policy.Expanding,
            self.name_label.sizePolicy().verticalPolicy(),
        )

        self.minus_button = QPushButton("-")
        self.minus_button.clicked.connect(self.decrease)
        self.minus_button.setProperty("class", "single_char_button")

        self.in_stock_edit = QLineEdit("ERROR")
        self.in_stock_edit.editingFinished.connect(self.update_in_stock)
        self.in_stock_edit.setSizePolicy(
            QSizePolicy.Policy.Minimum,
            self.in_stock_edit.sizePolicy().verticalPolicy(),
        )
        self.in_stock_edit.setProperty("class", "short_line_edit")

        self.wanted_label = QLabel("ERROR")

        self.plus_button = QPushButton("+")
        self.plus_button.clicked.connect(self.increase)
        self.plus_button.setProperty("class", "single_char_button")

        self.top_layout.addWidget(self.name_label)
        self.top_layout.addWidget(self.minus_button)
        self.top_layout.addWidget(self.in_stock_edit)
        self.top_layout.addWidget(self.wanted_label)
        self.top_layout.addWidget(self.plus_button)

        self.item_changed.connect(self.reload)
        self.reload()

    def _init_details(self) -> None:
        """details get initialized only when needed to speed up the app"""

        self.details = StorageFeedItemDetails(item=self.item)
        self.details.item_saved.connect(self.item_changed)
        self.details.item_saved.connect(self.toggle_details)
        self.details.item_deleted.connect(self.item_deleted)
        self.details.name_changed.connect(self.name_changed)
        self.details.setHidden(True)
        self.layout_.addWidget(self.details)

    def reload(self) -> None:
        self.name_label.setText(self.item.name)
        self.wanted_label.setText(f"/ {self.item.wanted} {self.item.unit}")
        self.in_stock_edit.setText(str(self.item.in_stock))
        self.minus_button.setEnabled(self.item.in_stock != 0)

        if self.details is not None:
            self.details.reload()

    def toggle_details(self) -> None:
        if self.details is None:
            self._init_details()
        self.details.setHidden(not self.details.isHidden())

    def increase(self) -> None:
        self.item.in_stock += 1
        self.on_changes()

    def decrease(self) -> None:
        try:
            self.item.in_stock -= 1
        except ValueError:
            self.item.in_stock = 0
        self.on_changes()

    def update_in_stock(self) -> None:
        try:
            new = float(self.in_stock_edit.text())

            if new == self.item.in_stock:
                return

            self.item.in_stock = new

        except ValueError as e:
            self.warn_message.emit(str(e))
            self.in_stock_edit.setText(str(self.item.in_stock))
            return

        self.on_changes()

    def on_changes(self) -> None:
        self.reload()
        self.item_changed.emit()


class CreateItemWizzard(QGroupBox):
    new_item = pyqtSignal(CartItem)

    def __init__(self) -> None:
        super().__init__()
        self.item = CartItem()
        self.init_ui()

    def init_ui(self) -> None:
        self.plus_button = QPushButton("+ Create Item +")
        self.plus_button.clicked.connect(self.toggle_details)

        self.details = StorageFeedItemDetails(self.item, is_for_wizzard=True)
        self.details.item_saved.connect(self.save_item)
        self.details.setHidden(True)

        plus_layout = QHBoxLayout()
        plus_layout.addStretch()
        plus_layout.addWidget(self.plus_button)
        plus_layout.addStretch()
        self.layout_ = QVBoxLayout()
        self.layout_.addLayout(plus_layout)
        self.layout_.addWidget(self.details)

        self.setLayout(self.layout_)

    def toggle_details(self) -> None:
        self.details.setHidden(not self.details.isHidden())
        if self.details.isHidden():
            self.plus_button.setText("+ Create Item +")
        else:
            self.plus_button.setText("- Collapse -")

    def save_item(self) -> None:
        self.new_item.emit(self.item)
        self.item = CartItem()
        self.details.item = self.item
        self.details.reload()
        self.toggle_details()


class StorageFeed(BaseFeed):
    item_changed = pyqtSignal(CartItem)
    item_deleted = pyqtSignal(CartItem)
    new_item = pyqtSignal(CartItem)
    name_changed = pyqtSignal(CartItem, str)
    warn_message = pyqtSignal(str)

    def init_ui(self) -> None:
        self.scroll_widget = QWidget()
        self.scroll_layout = QVBoxLayout()
        self.item_layout = QVBoxLayout()

        self.wizzard = CreateItemWizzard()
        self.wizzard.new_item.connect(self.new_item)

        self.scroll_layout.addWidget(self.wizzard)
        self.scroll_layout.addLayout(self.item_layout)
        self.scroll_layout.addStretch(1)

        self.scroll_widget.setLayout(self.scroll_layout)
        self.setWidget(self.scroll_widget)

        self.setWidgetResizable(True)
        self.setVerticalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOn
        )
        self.setHorizontalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        )
        QScroller.grabGesture(
            self.viewport(), QScroller.ScrollerGestureType.TouchGesture
        )

    def _create_ui_item(self, item: CartItem) -> StorageFeedItem:
        si = StorageFeedItem(item)
        si.item_changed.connect(lambda x=item: self.item_changed.emit(x))
        si.item_deleted.connect(lambda x=item: self.item_deleted.emit(x))
        si.name_changed.connect(lambda n, i=item: self.name_changed.emit(i, n))
        si.warn_message.connect(self.warn_message)
        return si


def main() -> None:
    app = QApplication(sys.argv)
    cart_items = [
        CartItem(
            name="Salt",
            wanted=1,
            in_stock=0,
            unit="Stk",
            only_once=False,
        ),
        CartItem(
            name="Potatoes",
            wanted=2.5,
            in_stock=1,
            unit="kg",
            only_once=False,
        ),
    ]
    for i in range(40):
        i += 1
        cart_items.append(
            CartItem(
                name=f"Item {i}",
                wanted=2 * i,
                in_stock=i,
                unit="Stk",
                only_once=False,
            )
        )

    storage_feed = StorageFeed(cart_items)
    storage_feed.item_changed.connect(print)
    storage_feed.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



#!/usr/bin/python3

from dataclasses import dataclass
from time import time
from PyQt6.QtWidgets import (
    QLabel,
    QVBoxLayout,
    QWidget,
)
from PyQt6.QtCore import QTimer


@dataclass
class ErrorBarError:
    message: str
    expiration_time: float = 5
    severity: str = "WARNING"

    @property
    def is_dead(self):
        return time() > self.expiration_time


class ErrorBar(QWidget):
    def __init__(self):
        super().__init__()
        self.errors = list()
        self.init_ui()
        self.init_timer()
        self.setHidden(True)

    def init_ui(self):
        self.layout_ = QVBoxLayout()
        self.setLayout(self.layout_)

    def init_timer(self) -> None:
        self.timer = QTimer()
        self.timer.timeout.connect(self.check_errors)
        self.timer.start(1000)

    def clear_layout(self) -> None:
        while not self.layout_.isEmpty():
            del_item = self.layout_.itemAt(0)
            self.layout_.removeItem(del_item)
            del_item.widget().deleteLater()

    def reload(self) -> None:
        if not self.errors and not self.isHidden():
            self.setHidden(True)

        self.clear_layout()
        for error in self.errors:
            lbl = QLabel(error.message)
            lbl.setWordWrap(True)
            lbl.setProperty("class", error.severity)
            self.layout_.addWidget(lbl)

    def show_error(self,
                   message: str,
                   ttl: float = 5,
                   ) -> None:
        self.show_generic_error(
            message=message,
            ttl=ttl,
            severity="ERROR",
        )

    def show_warning(self,
                     message: str,
                     ttl: float = 3.5,
                     ) -> None:
        self.show_generic_error(
            message=message,
            ttl=ttl,
            severity="WARNING",
        )

    def show_generic_error(self,
                           message: str,
                           severity: str,
                           ttl: float = 5,
                           ) -> None:

        self.errors.append(
            ErrorBarError(
                message=message,
                expiration_time=time() + ttl,
                severity=severity,
            )
        )
        self.setHidden(False)
        self.reload()

    def check_errors(self):
        if not self.errors:
            return
        for error in self.errors:
            if error.is_dead:
                self.errors.remove(error)
                self.reload()



#!/usr/bin/env python

"""
Source:
github.com/azsde/BatchMkvToolbox/blob/main/ui/customLayout/FlowLayout.py
#############################################################################
##
## Copyright (C) 2013 Riverbank Computing Limited.
## Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
## All rights reserved.
##
## This file is part of the examples of PyQt.
##
## $QT_BEGIN_LICENSE:BSD$
## You may use this file under the terms of the BSD license as follows:
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are
## met:
##   * Redistributions of source code must retain the above copyright
##     notice, this list of conditions and the following disclaimer.
##   * Redistributions in binary form must reproduce the above copyright
##     notice, this list of conditions and the following disclaimer in
##     the documentation and/or other materials provided with the
##     distribution.
##   * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor
##     the names of its contributors may be used to endorse or promote
##     products derived from this software without specific prior written
##     permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
## LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
## A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
## OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
## SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
## LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
## DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
## THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
## (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
## OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
## $QT_END_LICENSE$
##
#############################################################################
"""

from PyQt6.QtCore import QPoint, QRect, QSize, Qt, pyqtSignal
from PyQt6.QtWidgets import (QApplication, QLayout, QPushButton, QSizePolicy,
                             QWidget)


class Window(QWidget):
    def __init__(self):
        super(Window, self).__init__()

        flow_layout = FlowLayout()
        flow_layout.addWidget(QPushButton("Short"))
        flow_layout.addWidget(QPushButton("Longer"))
        flow_layout.addWidget(QPushButton("Different text"))
        flow_layout.addWidget(QPushButton("More text"))
        flow_layout.addWidget(QPushButton("Even longer button text"))
        self.setLayout(flow_layout)

        self.setWindowTitle("Flow Layout")


class FlowLayout(QLayout):
    widthChanged = pyqtSignal(int)

    def __init__(
            self,
            parent: QWidget = None,
            margin: int = 0,
            spacing: int = -1,
            orientation=Qt.Orientation.Horizontal,
            ) -> None:

        super(FlowLayout, self).__init__(parent)

        if parent is not None:
            self.setContentsMargins(margin, margin, margin, margin)

        self.setSpacing(spacing)
        self.item_list = []
        self.orientation = orientation

    def __del__(self) -> None:
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, item) -> None:
        self.item_list.append(item)

    def count(self) -> int:
        return len(self.item_list)

    def itemAt(self, index: int):
        if index >= 0 and index < len(self.item_list):
            return self.item_list[index]

        return None

    def takeAt(self, index):
        if index >= 0 and index < len(self.item_list):
            return self.item_list.pop(index)

        return None

    def expandingDirections(self):
        return Qt.Orientation(Qt.Orientation(0))

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        if (self.orientation == Qt.Orientation.Horizontal):
            return self.doLayoutHorizontal(QRect(0, 0, width, 0), True)
        elif (self.orientation == Qt.Orientation.Vertical):
            return self.doLayoutVertical(QRect(0, 0, width, 0), True)

    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        if (self.orientation == Qt.Orientation.Horizontal):
            self.doLayoutHorizontal(rect, False)
        elif (self.orientation == Qt.Orientation.Vertical):
            self.doLayoutVertical(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QSize()

        for item in self.item_list:
            size = size.expandedTo(item.minimumSize())

        margin, _, _, _ = self.getContentsMargins()

        size += QSize(2 * margin, 2 * margin)
        return size

    def doLayoutHorizontal(self, rect, testOnly):
        # Get initial coordinates of the drawing region (should be 0, 0)
        x = rect.x()
        y = rect.y()
        lineHeight = 0
        i = 0
        for item in self.item_list:
            wid = item.widget()
            # Space X and Y is item spacing horizontally and vertically
            spaceX = self.spacing() + wid.style().layoutSpacing(
                    QSizePolicy.ControlType.PushButton,
                    QSizePolicy.ControlType.PushButton,
                    Qt.Orientation.Horizontal
            )
            spaceY = self.spacing() + wid.style().layoutSpacing(
                    QSizePolicy.ControlType.PushButton,
                    QSizePolicy.ControlType.PushButton,
                    Qt.Orientation.Vertical
            )
            # Determine the coordinate we want to place the item at
            # It should be placed at: initial coordinate of the rect + width
            #                         of the item + spacing
            nextX = x + item.sizeHint().width() + spaceX
            # If the calculated nextX is greater than the outer bound...
            if nextX - spaceX > rect.right() and lineHeight > 0:
                x = rect.x()  # Reset X coordinate to origin of drawing region
                # Move Y coordinate to the next line
                y = y + lineHeight + spaceY
                # Recalculate nextX based on the new X coordinate
                nextX = x + item.sizeHint().width() + spaceX
                lineHeight = 0

            if not testOnly:
                item.setGeometry(QRect(QPoint(x, y), item.sizeHint()))

            x = nextX  # Store the next starting X coordinate for next item
            lineHeight = max(lineHeight, item.sizeHint().height())
            i = i + 1

        return y + lineHeight - rect.y()

    def doLayoutVertical(self, rect, testOnly):
        # Get initial coordinates of the drawing region (should be 0, 0)
        x = rect.x()
        y = rect.y()
        # Initalize column width and line height
        columnWidth = 0
        lineHeight = 0

        # Space between items
        spaceX = 0
        spaceY = 0

        # Variables that will represent the position of the
        # widgets in a 2D Array
        i = 0
        j = 0
        for item in self.item_list:
            wid = item.widget()
            # Space X and Y is item spacing horizontally and vertically
            spaceX = self.spacing() + wid.style().layoutSpacing(
                    QSizePolicy.ControlType.PushButton,
                    QSizePolicy.ControlType.PushButton,
                    Qt.Orientation.Horizontal
            )
            spaceY = self.spacing() + wid.style().layoutSpacing(
                    QSizePolicy.ControlType.PushButton,
                    QSizePolicy.ControlType.PushButton,
                    Qt.Orientation.Vertical
            )
            # Determine the coordinate we want to place the item at
            # It should be placed at: initial coordinate of the rect + width
            #                         of the item + spacing
            nextY = y + item.sizeHint().height() + spaceY
            # If the calculated nextY is greater than the outer bound, move to
            # the next column
            if nextY - spaceY > rect.bottom() and columnWidth > 0:
                y = rect.y()  # Reset y coordinate to origin of drawing region
                # Move X coordinate to the next column
                x = x + columnWidth + spaceX
                # Recalculate nextX based on the new X coordinate
                nextY = y + item.sizeHint().height() + spaceY
                # Reset the column width
                columnWidth = 0

                # Set indexes of the item for the 2D array
                j += 1
                i = 0

            # Assign 2D array indexes
            item.x_index = i
            item.y_index = j

            # Only call setGeometry (which place the actual widget using
            # coordinates) if testOnly is false
            # For some reason, Qt framework calls the doLayout methods with
            # testOnly set to true (WTF ??)
            if not testOnly:
                item.setGeometry(QRect(QPoint(x, y), item.sizeHint()))

            y = nextY  # Store the next starting Y coordinate for next item
            # Update the width of the column
            columnWidth = max(columnWidth, item.sizeHint().width())
            # Update the height of the line
            lineHeight = max(lineHeight, item.sizeHint().height())

            i += 1

        # Only call setGeometry (which place the actual widget using
        # coordinates) if testOnly is false
        # For some reason, Qt framework calls the doLayout methods with
        # testOnly set to true (WTF ??)
        if not testOnly:
            self.calculateMaxWidth(i)
            # print("Total width : " + str(totalWidth))
            self.widthChanged.emit(
                self.totalMaxWidth + spaceX * self.itemsOnWidestRow
            )
            # self.widthChanged.emit(self.totalMaxWidth)
        return lineHeight

    # Method to calculate the maximum width among each "row" of the flow layout
    # This will be useful to let the UI know the total width of the flow layout
    def calculateMaxWidth(self, numberOfRows):
        # Init variables
        self.totalMaxWidth = 0
        self.itemsOnWidestRow = 0

        # For each "row", calculate the total width by adding the width of
        # each item and then update the totalMaxWidth if the calculated width
        # is greater than the current value
        # Also update the number of items on the widest row
        for i in range(numberOfRows):
            rowWidth = 0
            itemsOnWidestRow = 0
            for item in self.item_list:
                # Only compare items from the same row
                if (item.x_index == i):
                    rowWidth += item.sizeHint().width()
                    itemsOnWidestRow += 1
                if (rowWidth > self.totalMaxWidth):
                    self.totalMaxWidth = rowWidth
                    self.itemsOnWidestRow = itemsOnWidestRow


if __name__ == '__main__':
    import sys

    app = QApplication(sys.argv)
    mainWin = Window()
    mainWin.show()
    sys.exit(app.exec())



#!/usr/bin/python3
# -*- coding: utf-8 -*-

import os
from pathlib import Path
import json
from dataclasses import dataclass, field
from time import time


@dataclass
class CartItem:
    name: str = ""
    _name: str = field(init=False, repr=False, default="")
    wanted: float
    _wanted: float = field(init=False, repr=False, default=1)
    in_stock: float
    _in_stock: float = field(init=False, repr=False, default=0)
    unit: str = ""
    only_once: bool = False
    tags: list[str]
    _tags: list[str] = field(init=False, repr=False)
    mtime: float | None = None
    hidden_temporary: bool = False

    def __post_init__(self) -> None:
        if self.mtime is None:
            self._on_change()

    def __dict__(self) -> dict:
        return {
            "name": self.name,
            "wanted": self.wanted,
            "in_stock": self.in_stock,
            "unit": self.unit,
            "only_once": self.only_once,
            "tags": self.tags,
            "mtime": self.mtime,
        }

    @classmethod
    def from_dict(cls, d) -> "CartItem":
        return cls(
            name=d["name"],
            wanted=d["wanted"],
            in_stock=d["in_stock"],
            unit=d["unit"],
            only_once=d.get("only_once", False),
            tags=d.get("tags", list()),
            mtime=d.get("mtime", 0),
        )

    def __eq__(self, other) -> bool:
        """Attention: Ignores mtime!"""
        if self.name != other.name:
            return False
        if self.wanted != other.wanted:
            return False
        if self.in_stock != other.in_stock:
            return False
        if self.unit != other.unit:
            return False
        if self.only_once != other.only_once:
            return False
        if self.tags != other.tags:
            return False
        if self.hidden_temporary != other.hidden_temporary:
            return False
        return True

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, new: str) -> None:
        if type(new) is property:
            # initial value not specified, use default
            self._name = CartItem._name
            return

        if type(new) != str:
            raise ValueError()

        if self._name == new:
            return

        self._name = new
        self._on_change()

    @property
    def in_stock(self) -> float | int:
        return self._int_if_possible(self._in_stock)

    @in_stock.setter
    def in_stock(self, new: any) -> None:
        if type(new) is property:
            # initial value not specified, use default
            self._in_stock = CartItem._in_stock
            return

        new = self._to_float(new)

        if new < 0:
            raise ValueError("Negative numbers are not supported!")

        if new == self.in_stock:
            return

        self._in_stock = new
        self._on_change()

    @property
    def wanted(self) -> float | int:
        return self._int_if_possible(self._wanted)

    @wanted.setter
    def wanted(self, new):
        if type(new) is property:
            # initial value not specified, use default
            self._wanted = CartItem._wanted
            return

        new = self._to_float(new)

        if self._wanted == new:
            return

        if new <= 0:
            raise ValueError("You can not want nothing or less than nothing")

        self._wanted = new
        self._on_change()

    @property
    def missing(self) -> float | int:
        if self.wanted < self.in_stock:
            return 0
        return self._int_if_possible(self.wanted - self.in_stock)

    @missing.setter
    def missing(self, new) -> None:
        new = self._to_float(new)

        if self.wanted < new:
            self.in_stock = 0
            return

        self.in_stock = self.wanted - new

    @property
    def needs_restock(self) -> bool:
        return self.wanted > self.in_stock

    @property
    def is_dead(self) -> bool:
        return self.only_once and not self.needs_restock

    @property
    def tags(self) -> list[str]:
        """There is a problem with mutable properties like lists:
        If you use e.g. item.tags.append() the setter does not get called.
        To at least make this a more obvious bug we return a copy of the list
        using [:] so .append() will not work. """
        return self._tags[:]

    @tags.setter
    def tags(self, new) -> None:
        if type(new) is property:
            # initial value not specified, use default
            self._tags = []
            return

        if getattr(self, "_tags", None) == new:
            return

        self._tags = new
        self._on_change()

    def add_tag(self, name) -> None:
        self._tags.append(name)
        self._on_change()

    def remove_tag(self, name) -> None:
        self._tags.remove(name)
        self._on_change()

    def _on_change(self) -> None:
        self.mtime = time()

    def bought_all(self) -> None:
        self.in_stock = self.wanted

    def _int_if_possible(self, value: float) -> float | int:
        """str(1.0) looks worse than str(1) so we convert if possible"""

        # float is not always exact and we don't need more than 6 digits
        value = round(value, 6)

        if int(value) == value:
            return int(value)
        return value

    def _to_float(self, value: any) -> float:
        if type(value) is str:
            if value == "":
                float(value)  # raises exception
            if value[-1] == ".":
                raise ValueError("String ends with a dot")
        return float(value)


class DuplicateError(Exception):
    pass


class FileManager:
    app_name: str = "karoto"
    list_extension: str = ".json"

    def _get_list_path(self, name: str, full_path: str = None) -> Path:
        if full_path is not None:
            return Path(full_path)
        elif name is not None:
            xdg_data_home = os.environ.get('XDG_DATA_HOME') or \
                Path.home() / '.local' / 'share'
            return Path(xdg_data_home) / self.app_name / (name +
                                                          self.list_extension)
        else:
            return self.list_path

    def load_list(self, name: str = "default", full_path: str = None) -> list:
        self.list_path = self._get_list_path(name=name, full_path=full_path)

        storage_dict = json.load(open(self.list_path))

        ret = list()
        for d_item in storage_dict["items"]:
            item = CartItem.from_dict(d_item)
            ret.append(item)
        return ret

    def _storage_to_dict(self, storage: list) -> dict:
        d = dict(items=list())
        for item in storage:
            d["items"].append(item.__dict__())
        return d

    def save_list(self,
                  storage: list,
                  name: str = None,
                  full_path: str = None,
                  ) -> None:
        list_path = self._get_list_path(name=name, full_path=full_path)

        storage = self._storage_to_dict(storage)

        list_path.parent.mkdir(parents=True, exist_ok=True)
        json.dump(storage, open(list_path, "w"), indent=4)


class Backend:
    def __init__(self,
                 list_name: str = "defaut",
                 list_file: str = None,
                 ) -> None:
        super().__init__()
        self.fm = FileManager()

        self.load_list(list_name=list_name, full_path=list_file)

    def load_list(self, list_name: str = None, full_path: str = None) -> None:
        try:
            self.storage = self.fm.load_list(
                name=list_name,
                full_path=full_path,
            )
        except FileNotFoundError:
            self.storage = list()

    def save_list(self, list_name: str = None, full_path: str = None) -> None:
        self.fm.save_list(
            self.storage,
            name=list_name,
            full_path=full_path,
        )

    def _get_filtered_items(
            self,
            search: str = "",
            filter_tags: list[str] = [],
            ) -> list:

        if search == "" and filter_tags == []:
            ret = self.storage
            ret.sort(key=lambda k: k.mtime, reverse=True)
            return ret

        new_list = list()
        for item in self.storage:
            if search != "" and item.name.lower().find(search.lower()) == -1:
                continue

            found_missing = False
            for tag in filter_tags:
                if tag not in item.tags:
                    found_missing = True
                    break
            if not found_missing:
                new_list.append(item)

        new_list.sort(key=lambda k: k.mtime, reverse=True)
        return new_list

    def get_items_in_stock(
            self,
            search: str = "",
            filter_tags: list[str] = [],
            ) -> list:

        return self._get_filtered_items(search=search, filter_tags=filter_tags)

    def get_items_to_buy(
            self,
            search: str = "",
            filter_tags: list[str] = [],
            ) -> list:

        items = list()
        for item in self._get_filtered_items(
                search=search,
                filter_tags=filter_tags,
                ):
            if item.needs_restock and not item.hidden_temporary:
                items.append(item)
        return items

    def get_tags(
            self,
            search: str = "",
            filter_tags: list[str] = [],
            storage_mode=True,
            ) -> list[tuple[str, int]]:
        if storage_mode:
            all_items = self.get_items_in_stock(
                search=search,
                filter_tags=filter_tags,
            )
        else:
            all_items = self.get_items_to_buy(
                search=search,
                filter_tags=filter_tags,
            )

        all_tags = []
        for item in all_items:
            all_tags += item.tags

        counted_tags = []
        for tag in set(all_tags):
            counted_tags.append((tag, all_tags.count(tag)))

        return counted_tags

    def add_item(self, item: CartItem):
        if self.get_item_by_name(item.name) is not None:
            raise DuplicateError("A item with this name already exists")
        self.storage.append(item)

    def delete(self, item: CartItem) -> None:
        self.storage.remove(item)

    def get_item_by_name(self, name: str) -> CartItem:
        for item in self.storage:
            if name == item.name:
                return item



from PyQt6.QtWidgets import QLabel
from PyQt6.QtCore import pyqtSignal
from time import time


class ClickableLabel(QLabel):
    clicked = pyqtSignal()
    _last_press = 0

    @property
    def _press_duration(self) -> float:
        return time() - self._last_press

    def mousePressEvent(self, event):
        self._last_press = time()
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        if self._press_duration < 0.2:  # TODO: steal value from QPushButton?
            self.clicked.emit()

        super().mouseReleaseEvent(event)



_item_height = 18
_sixty = "white"
_thirty = "#fdc033"
_ten = "#80e241"

default = f"""
* {{
    background-color: {_sixty};
    font-size: 14px;
    color: black;
}}

QPushButton {{
    height: {_item_height}px;
    background-color: {_ten};
    border: 1px solid grey;
    border-radius: {_item_height/2}px;
    padding-top: 3px;
    padding-bottom: 3px;
    padding-left: 10px;
    padding-right: 10px;
}}

QLineEdit {{
    border-radius: 10px;
    padding: 2px;
    background-color: {_sixty};
}}

QLabel, QCheckBox {{
    background-color: {_thirty};
}}

.top_bar {{
    min-height: 40px;
    qproperty-iconSize: 40px;
}}

#error_bar QLabel {{
    min-height: 40px;
    font-weight: bold;
    padding-top: 3px;
    padding-bottom: 3px;
    padding-left: 10px;
    padding-right: 10px;
}}

.WARNING {{
    background-color: yellow;
}}

.ERROR {{
    background-color: red;
}}

#menu_button {{
    min-width: 60px;
}}

#search_field {{
    padding-left: 6px;
    border: 1px solid grey;
}}

QGroupBox {{
    background-color: {_thirty};
    border-radius: 6px;
}}

.short_line_edit {{
    width: 40px;
}}

.single_char_button {{
    padding: 2px;
    border: 1px solid grey;
    border-radius: {_item_height/2}px;
    width: {_item_height}px;
    height: {_item_height}px;
    font-weight: bold;
}}

QPushButton:disabled {{
    border: 1px solid #77AA77;
    background-color: #88BB88;
    color: #777777;
}}
"""

# Source: https://tabler-icons.io/ search for "filter"
filter_plus_icon_svg = b"""
<svg xmlns="http://www.w3.org/2000/svg" class="icon icon-tabler
icon-tabler-filter-plus" width="24" height="24" viewBox="0 0 24 24"
stroke-width="2" stroke="currentColor" fill="none" stroke-linecap="round"
stroke-linejoin="round">
   <path stroke="none" d="M0 0h24v24H0z" fill="none"></path>
   <path d="M12 20l-3 1v-8.5l-4.48 -4.928a2 2 0 0 1 -.52
   -1.345v-2.227h16v2.172a2 2 0 0 1 -.586 1.414l-4.414 4.414v3"></path>
   <path d="M16 19h6"></path>
   <path d="M19 16v6"></path>
</svg>
"""

filter_minus_icon_svg = b"""
<svg xmlns="http://www.w3.org/2000/svg" class="icon icon-tabler
icon-tabler-filter-minus" width="24" height="24" viewBox="0 0 24 24"
stroke-width="2" stroke="currentColor" fill="none" stroke-linecap="round"
stroke-linejoin="round">
   <path stroke="none" d="M0 0h24v24H0z" fill="none"></path>
   <path d="M12 20l-3 1v-8.5l-4.48 -4.928a2 2 0 0 1 -.52
   -1.345v-2.227h16v2.172a2 2 0 0 1 -.586 1.414l-4.414 4.414v3"></path>
   <path d="M16 19h6"></path>
</svg>
"""

# Source: https://feathericons.com/?query=cart and
#         https://feathericons.com/?query=home (modified with Inkscape)
menu_home_icon_svg = b"""
<svg
   width="50"
   height="24"
   viewBox="0 0 13.229168 6.3499998"
   version="1.1"
   id="svg5"
   xml:space="preserve"
   sodipodi:docname="home.svg"
   inkscape:version="1.2.2 (b0a8486541, 2022-12-01)"
   xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape"
   xmlns:sodipodi="http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"
   xmlns="http://www.w3.org/2000/svg"
   xmlns:svg="http://www.w3.org/2000/svg"><sodipodi:namedview
     id="namedview7"
     pagecolor="#ffffff"
     bordercolor="#666666"
     borderopacity="1.0"
     inkscape:showpageshadow="2"
     inkscape:pageopacity="0.0"
     inkscape:pagecheckerboard="0"
     inkscape:deskcolor="#d1d1d1"
     inkscape:document-units="mm"
     showgrid="false"
     inkscape:zoom="8.1933293"
     inkscape:cx="22.823445"
     inkscape:cy="20.809612"
     inkscape:window-width="1362"
     inkscape:window-height="744"
     inkscape:window-x="0"
     inkscape:window-y="0"
     inkscape:window-maximized="1"
     inkscape:current-layer="layer1" /><defs
     id="defs2" /><g
     inkscape:label="Ebene 1"
     inkscape:groupmode="layer"
     id="layer1"><g
       style="fill:none;fill-opacity:1;stroke:#535353;stroke-width:2;stroke-linecap:round;stroke-linejoin:round;stroke-opacity:1"
       id="g342"
       transform="matrix(0.17637215,0,0,0.15656522,0,1.5875)"><circle
         cx="9"
         cy="21"
         r="1"
         id="circle327"
         style="fill:none;fill-opacity:1;stroke:#535353;stroke-opacity:1" />
       <circle
         cx="20"
         cy="21"
         r="1"
         id="circle329"
         style="fill:none;fill-opacity:1;stroke:#535353;stroke-opacity:1" />
       <path
         d="m 1,1 h 4 l 2.68,13.39 a 2,2 0 0 0 2,1.61 h 9.72 a 2,2 0 0 0 2,
         -1.61 L 23,6 H 6"
         id="path331"
         style="fill:none;fill-opacity:1;stroke:#535353;stroke-opacity:1" />
           </g><g
       style="fill:none;stroke:currentColor;stroke-width:2;
       stroke-linecap:round;stroke-linejoin:round"
       id="g958"
       transform="matrix(0.28863671,0,0,0.28863671,6.8575182,-0.28864441)">
       <path
         d="m 3,9 9,-7 9,7 v 11 a 2,2 0 0 1 -2,2 H 5 A 2,2 0 0 1 3,20 Z"
         id="path946" /><polyline
         points="9 22 9 12 15 12 15 22"
         id="polyline948" /></g></g></svg>
"""

menu_cart_icon_svg = b"""
<svg
   width="50"
   height="24"
   viewBox="0 0 13.229168 6.3499998"
   version="1.1"
   id="svg5"
   xml:space="preserve"
   sodipodi:docname="cart.svg"
   inkscape:version="1.2.2 (b0a8486541, 2022-12-01)"
   xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape"
   xmlns:sodipodi="http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"
   xmlns="http://www.w3.org/2000/svg"
   xmlns:svg="http://www.w3.org/2000/svg"><sodipodi:namedview
     id="namedview7"
     pagecolor="#ffffff"
     bordercolor="#666666"
     borderopacity="1.0"
     inkscape:showpageshadow="2"
     inkscape:pageopacity="0.0"
     inkscape:pagecheckerboard="0"
     inkscape:deskcolor="#d1d1d1"
     inkscape:document-units="mm"
     showgrid="false"
     inkscape:zoom="8.1933293"
     inkscape:cx="22.88447"
     inkscape:cy="20.687561"
     inkscape:window-width="1362"
     inkscape:window-height="744"
     inkscape:window-x="0"
     inkscape:window-y="0"
     inkscape:window-maximized="1"
     inkscape:current-layer="layer1" /><defs
     id="defs2" /><g
     inkscape:label="Ebene 1"
     inkscape:groupmode="layer"
     id="layer1"><g
       style="fill:none;fill-opacity:1;stroke:#000000;stroke-width:2;
         stroke-linecap:round;stroke-linejoin:round;stroke-opacity:1"
       id="g342"
       transform="scale(0.31101448,0.27608696)"><circle
         cx="9"
         cy="21"
         r="1"
         id="circle327"
         style="fill:none;fill-opacity:1;stroke:#000000;stroke-opacity:1" />
       <circle
         cx="20"
         cy="21"
         r="1"
         id="circle329"
         style="fill:none;fill-opacity:1;stroke:#000000;stroke-opacity:1" />
       <path
         d="m 1,1 h 4 l 2.68,13.39 a 2,2 0 0 0 2,1.61 h 9.72 a 2,2 0 0 0 2,
         -1.61 L 23,6 H 6"
         id="path331"
         style="fill:none;fill-opacity:1;stroke:#000000;stroke-opacity:1" />
           </g><g
       style="fill:none;stroke:#545454;stroke-width:2;stroke-linecap:round;
         stroke-linejoin:round;stroke-opacity:1"
       id="g958"
       transform="matrix(0.19242448,0,0,0.19242448,8.981401,0.81298705)">
       <path
         d="m 3,9 9,-7 9,7 v 11 a 2,2 0 0 1 -2,2 H 5 A 2,2 0 0 1 3,20 Z"
         id="path946"
         style="stroke:#545454;stroke-opacity:1" /><polyline
         points="9 22 9 12 15 12 15 22"
         id="polyline948"
         style="stroke:#545454;stroke-opacity:1" /></g></g></svg>
"""
